/* This file is part of Clementine.
   Copyright 2012, David Sansome <me@davidsansome.com>

   Clementine is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Clementine is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Clementine.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef GOOGLEDRIVESTREAM_H
#define GOOGLEDRIVESTREAM_H

#include "clementine-config.h"

#include <QList>
#include <QNetworkAccessManager>
#include <QObject>
#include <QSslError>
#include <QUrl>

#include <google/sparsetable>
#include <taglib/tiostream.h>

class CloudStream : public QObject, public TagLib::IOStream {
  Q_OBJECT
 public:
  CloudStream(const QUrl& url,
              const QString& filename,
#ifdef HAVE_TAGLIB2
              const size_t length,
#else
              const long length,
#endif
              const QString& auth);

  // Taglib::IOStream
  virtual TagLib::FileName name() const;
#ifdef HAVE_TAGLIB2
  virtual TagLib::ByteVector readBlock(size_t length) override;
  virtual void insert(const TagLib::ByteVector&, TagLib::offset_t, size_t) override;
  virtual void removeBlock(TagLib::offset_t, size_t) override;
  virtual void seek(TagLib::offset_t offset, TagLib::IOStream::Position p) override;
  virtual void truncate(TagLib::offset_t) override;
  virtual TagLib::offset_t tell() const override;
  virtual TagLib::offset_t length() override;
#else
  virtual TagLib::ByteVector readBlock(ulong length) override;
  virtual void insert(const TagLib::ByteVector&, ulong, ulong) override;
  virtual void removeBlock(ulong, ulong);
  virtual void seek(long offset, TagLib::IOStream::Position p) override;
  virtual void truncate(long) override;
  virtual long tell() const override;
  virtual long length() override;
#endif
  virtual void writeBlock(const TagLib::ByteVector&) override;
  virtual bool readOnly() const override;
  virtual bool isOpen() const override;
  virtual void clear() override;

  google::sparsetable<char>::size_type cached_bytes() const {
    return cache_.num_nonempty();
  }

  int num_requests() const { return num_requests_; }

  // Use educated guess to request the bytes that TagLib will probably want.
  void Precache();

 private:
  bool CheckCache(int start, int end);
  void FillCache(int start, TagLib::ByteVector data);
  TagLib::ByteVector GetCached(int start, int end);

 private slots:
  void SSLErrors(const QList<QSslError>& errors);

 private:
  const QUrl url_;
  const QString filename_;
  const QByteArray encoded_filename_;
#ifdef HAVE_TAGLIB2
  const size_t length_;
#else
  const ulong length_;
#endif
  const QString auth_;

  int cursor_;
  std::unique_ptr<QNetworkAccessManager> network_;

  google::sparsetable<char> cache_;
  int num_requests_;
};

#endif  // GOOGLEDRIVESTREAM_H
